#ifndef TYPON_CORE_CONTINUATION_HPP_INCLUDED
#define TYPON_CORE_CONTINUATION_HPP_INCLUDED

#include <atomic>
#include <concepts>
#include <coroutine>
#include <cstdint>
#include <exception>
#include <limits>


namespace typon
{

  struct ForkList
  {
    struct Node
    {
      Node * _next;
      std::coroutine_handle<> _coroutine;
      std::exception_ptr * _exception;
    };

    Node * _first = nullptr;

    template <typename Promise>
      requires requires (Promise p)
      {
        { p._exception } -> std::same_as<std::exception_ptr&>;
        { p._node } -> std::same_as<Node&>;
      }
    void insert(std::coroutine_handle<Promise> coroutine) noexcept
    {
      std::exception_ptr * exception = &(coroutine.promise()._exception);
      coroutine.promise()._node = { _first, coroutine, exception };
      _first = &(coroutine.promise()._node);
    }

    void check_exceptions()
    {
      for (auto node = _first; node != nullptr; node = node->_next)
      {
        std::exception_ptr & exception = *(node->_exception);
        if (exception)
        {
          std::rethrow_exception(exception);
        }
      }
    }

    void clear() noexcept
    {
      auto next = _first;
      while(next)
      {
        auto node = next;
        next = next->_next;
        node->_coroutine.destroy();
      }
      _first = nullptr;
    }
  };


  struct Continuation
  {
    struct Data
    {
      using u64 = std::uint_fast64_t;

      static constexpr u64 UMAX = std::numeric_limits<u64>::max();

      std::coroutine_handle<> _coroutine;
      std::coroutine_handle<> _continuation;

      ForkList _children;

      std::atomic<u64> _n = UMAX;
      u64 _thefts = 0;

      Data(std::coroutine_handle<> coroutine) noexcept
        : _coroutine(coroutine)
      {}
    };

    Data * _data;

    Continuation() noexcept {}

    template <typename Promise>
    Continuation(std::coroutine_handle<Promise> coroutine) noexcept
      : _data(&(coroutine.promise()._data))
    {}

    auto & thefts() noexcept
    {
      return _data->_thefts;
    }

    auto & children() noexcept
    {
      return _data->_children;
    }

    auto & n() noexcept
    {
      return _data->_n;
    }

    void resume()
    {
      _data->_coroutine.resume();
    }

    operator std::coroutine_handle<>() noexcept
    {
      return _data->_coroutine;
    }

    std::coroutine_handle<> continuation() noexcept
    {
      if (_data->_coroutine.done())
      {
        return _data->_continuation;
      }
      return _data->_coroutine;
    }
  };

}


#endif // TYPON_CONTINUATION_HPP_INCLUDED
