#ifndef TYPON_CORE_JOIN_HPP_INCLUDED
#define TYPON_CORE_JOIN_HPP_INCLUDED

#include <coroutine>
#include <utility>

#include <typon/core/continuation.hpp>
#include <typon/core/result.hpp>


namespace typon
{

  struct [[nodiscard]] Sync {};


  template <typename T = void>
  struct [[nodiscard]] Join
  {
    struct promise_type;

    std::coroutine_handle<promise_type> _coroutine;

    Join(std::coroutine_handle<promise_type> coroutine) noexcept : _coroutine(coroutine) {}

    Join(const Join &) = delete;
    Join & operator=(const Join &) = delete;

    Join(Join && other) noexcept
      : _coroutine(std::exchange(other._coroutine, nullptr))
    {}

    Join & operator=(Join other)
    {
      std::swap(_coroutine, other._coroutine);
      return *this;
    }

    ~Join()
    {
      if (_coroutine)
      {
        _coroutine.promise()._data._children.clear();
        _coroutine.destroy();
      }
    }

    struct promise_type : Result<T>
    {
      using u64 = Continuation::Data::u64;
      static constexpr u64 UMAX = Continuation::Data::UMAX;

      Continuation::Data _data;

      promise_type() noexcept
        : _data(std::coroutine_handle<promise_type>::from_promise(*this))
      {}

      Join get_return_object() noexcept
      {
        return { std::coroutine_handle<promise_type>::from_promise(*this) };
      }

      std::suspend_always initial_suspend() noexcept
      {
        return {};
      }

      template <typename U>
      decltype(auto) await_transform(U && expr) noexcept
      {
        return std::forward<U>(expr);
      }

      auto await_transform(Sync &&) noexcept
      {
        struct awaitable
        {
          Continuation::Data & _data;

          bool await_ready() noexcept
          {
            if (u64 thefts = _data._thefts)
            {
              u64 n = _data._n.load(std::memory_order_acquire);
              if (n - (UMAX - thefts) == 0)
              {
                return true;
              }
              return false;
            }
            return true;
          }

          std::coroutine_handle<> await_suspend(std::coroutine_handle<promise_type> coroutine) noexcept
          {
            u64 thefts = _data._thefts;
            u64 n = _data._n.fetch_sub(UMAX - thefts, std::memory_order_acq_rel);
            if (n - (UMAX - thefts) == 0)
            {
              return coroutine;
            }
            return std::noop_coroutine();
          }

          void await_resume() noexcept
          {
            _data._thefts = 0;
            _data._n.store(UMAX, std::memory_order_release);
          }
        };

        return awaitable { _data };
      }

      auto final_suspend() noexcept
      {
        struct awaitable : std::suspend_always
        {
          std::coroutine_handle<> await_suspend(std::coroutine_handle<promise_type> coroutine) noexcept
          {
            Continuation::Data & data = coroutine.promise()._data;
            u64 thefts = data._thefts;
            u64 n = data._n.fetch_sub(UMAX - thefts, std::memory_order_acq_rel);
            if (n - (UMAX - thefts) == 0)
            {
              return data._continuation;
            }
            return std::noop_coroutine();
          }
        };

        return awaitable {};
      }
    };

    auto operator co_await() && noexcept
    {
      struct awaitable
      {
        std::coroutine_handle<promise_type> _coroutine;

        bool await_ready() noexcept
        {
          return false;
        }

        std::coroutine_handle<> await_suspend(std::coroutine_handle<> continuation) noexcept
        {
          _coroutine.promise()._data._continuation = continuation;
          return _coroutine;
        }

        decltype(auto) await_resume()
        {
          _coroutine.promise()._data._children.check_exceptions();
          return _coroutine.promise().get();
        }
      };

      return awaitable { _coroutine };
    }
  };

}

#endif // TYPON_JOIN_HPP_INCLUDED
