#ifndef TYPON_FUNDAMENTAL_DEQUE_HPP_INCLUDED
#define TYPON_FUNDAMENTAL_DEQUE_HPP_INCLUDED

#include <atomic>
#include <cstdint>
#include <memory>
#include <type_traits>
#include <utility>

#include <typon/fundamental/optional.hpp>
#include <typon/fundamental/ring_buffer.hpp>
#include <typon/fundamental/scope.hpp>


namespace typon::fdt::lock_free
{

  template <typename T>
  struct deque
  {
    using array_type = ring_buffer<T>;
    using pop_type = fdt::optional<T>;

    using u8 = typename ring_buffer<T>::u8;
    using u64 = typename ring_buffer<T>::u64;

    static constexpr typename pop_type::template state<1> Abort {};
    static constexpr typename pop_type::template state<2> Prune {};

    using enum std::memory_order;

    const u8 _bits;
    std::atomic<u64> _top {1};
    std::atomic<u64> _bottom {1};
    std::atomic<array_type *> _array;

    deque(u8 bits = 3) noexcept
      : _bits(bits)
      , _array(new array_type(bits))
    {}

    ~deque()
    {
      delete _array.load(relaxed);
    }

    void push(T x) noexcept
    {
      u64 bottom = _bottom.load(relaxed);
      u64 top = _top.load(acquire);
      array_type * array = _array.load(relaxed);
      if (bottom - top > array->capacity() - 1)
      {
        array = array->grow(top, bottom);
        _array.store(array);
      }
      array->put(bottom, x);
      std::atomic_thread_fence(release);
      _bottom.store(bottom + 1, relaxed);
    }

    pop_type pop() noexcept
    {
      u64 bottom = _bottom.load(relaxed) - 1;
      array_type * array = _array.load(relaxed);
      _bottom.store(bottom, relaxed);
      std::atomic_thread_fence(seq_cst);
      u64 top = _top.load(relaxed);
      pop_type x {};
      if (top <= bottom)
      {
        x = array->get(bottom);
        if (top == bottom)
        {
          if (!_top.compare_exchange_strong(top, top + 1, seq_cst, relaxed))
          {
            x = {};
          }
          _bottom.store(bottom + 1, relaxed);
        }
      }
      else
      {
        _bottom.store(bottom + 1, relaxed);
      }
      u64 capacity = array->capacity();
      if (capacity > u64(1) << _bits && capacity > (bottom - top) * 4)
      {
        x.set_state(Prune);
      }
      return x;
    }

    array_type * shrink() noexcept
    {
      u64 bottom = _bottom.load(relaxed);
      array_type * array = _array.load(relaxed);
      u64 top = _top.load(relaxed);
      array_type * next = array->shrink(top, bottom);
      if (next)
      {
        _array.store(next, relaxed);
        return array;
      }
      return nullptr;
    }

    pop_type steal() noexcept
    {
      u64 top = _top.load(acquire);
      std::atomic_thread_fence(seq_cst);
      u64 bottom = _bottom.load(acquire);
      if (top < bottom)
      {
        array_type * array = _array.load(consume);
        T x = array->get(top);
        if (!_top.compare_exchange_strong(top, top + 1, seq_cst, relaxed))
        {
          return { Abort };
        }
        return { x };
      }
      return {};
    }
  };

}


#endif // TYPON_FUNDAMENTAL_DEQUE_HPP_INCLUDED
