#ifndef TYPON_CORE_SCHEDULER_HPP_INCLUDED
#define TYPON_CORE_SCHEDULER_HPP_INCLUDED

#include <atomic>
#include <cstdint>
#include <exception>
#include <thread>
#include <vector>

#include <typon/fundamental/deque.hpp>
#include <typon/fundamental/event_count.hpp>
#include <typon/fundamental/gc.hpp>
#include <typon/fundamental/optional.hpp>
#include <typon/fundamental/random.hpp>

#include <typon/core/continuation.hpp>


namespace typon
{

  struct Scheduler
  {
    using uint = unsigned int;
    using Deque = fdt::lock_free::deque<Continuation>;
    using Task = typename Deque::pop_type;
    using GC = fdt::lock_free::gc;

    static inline thread_local uint thread_id;

    static Scheduler & get() noexcept
    {
      static Scheduler scheduler {std::thread::hardware_concurrency()};
      return scheduler;
    }

    static void push(Continuation task) noexcept
    {
      get()._deque[thread_id].push(task);
    }

    static void schedule(Continuation task) noexcept
    {
      Scheduler & scheduler = get();
      scheduler._deque[thread_id].push(task);
      scheduler._notifyer.notify_one();
    }

    static Task pop() noexcept
    {
      Scheduler & scheduler = get();
      Deque & deque = scheduler._deque[thread_id];
      Task task = deque.pop();
      if (task.match(Deque::Prune))
      {
        if (auto array = deque.shrink())
        {
          scheduler._gc.retire(array);
        }
      }
      return task;
    }

    std::atomic<uint> _actives = 0;
    std::atomic<uint> _thieves = 0;
    std::vector<Deque> _deque;
    std::vector<std::thread> _thread;
    std::atomic_bool _done {false};
    fdt::lock_free::event_count<> _notifyer;
    const uint _concurrency;
    GC _gc;

    Scheduler(uint concurrency) noexcept
      : _deque(concurrency + 1)
      , _concurrency(concurrency)
      , _gc(concurrency)
    {
      thread_id = concurrency;

      for (uint id = 0; id < concurrency; id++)
      {
        _thread.emplace_back([this, id]() {
          thread_id = id;

          Task task {};
          for(;;)
          {
            exploit_task(task);
            if (!wait_for_task(task))
            {
              break;
            }
          }
        });
      }
    }

    ~Scheduler() noexcept
    {
      _done.store(true);
      _notifyer.notify_all();
      for (auto & t : _thread)
      {
        t.join();
      }
    }

    void exploit_task(Task & task) noexcept
    {
      if (task)
      {
        if (_actives.fetch_add(1) == 0)
        {
          if (_thieves.load() == 0)
          {
            _notifyer.notify_one();
          }
        }
        while(task)
        {
          task->resume();
          task = _deque[thread_id].pop();
          if (task)
          {
            printf("[%u] ERROR unexpectedly pop valid task %p\n", thread_id, task->_data);
          }
        }
        _actives.fetch_sub(1);
      }
    }

    void explore_task(Task & task) noexcept
    {
      _gc.enter(thread_id);
      for (uint i = 0; i < _concurrency * 2 + 1; i++)
      {
        uint id = fdt::random::random() % _concurrency;
        if (id == thread_id)
        {
          task = _deque.back().steal();
        }
        else
        {
          task = _deque[id].steal();
        }
        if (task)
        {
          task->thefts()++;
          break;
        }
      }
      _gc.leave(thread_id);
    }

    bool wait_for_task(Task & task) noexcept
    {
    wait_for_task:
      _thieves.fetch_add(1);
    explore_task:
      explore_task(task);
      if (task)
      {
        if (_thieves.fetch_sub(1) == 1)
        {
          _notifyer.notify_one();
        }
        return true;
      }
      auto key = _notifyer.prepare_wait();
      task = _deque.back().steal();
      if (task || task.match(Deque::Abort))
      {
        _notifyer.cancel_wait();
        if (task)
        {
          if (_thieves.fetch_sub(1) == 1)
          {
            _notifyer.notify_one();
          }
          return true;
        }
        goto explore_task;
      }
      if (_done.load())
      {
        _notifyer.cancel_wait();
        _notifyer.notify_all();
        _thieves.fetch_sub(1);
        return false;
      }
      if (_thieves.fetch_sub(1) == 1)
      {
        if (_actives.load() > 0)
        {
          _notifyer.cancel_wait();
          goto wait_for_task;
        }
      }
      _notifyer.wait(key);
      return true;
    }
  };

}

#endif // TYPON_SCHEDULER_HPP_INCLUDED
